# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import absolute_import
import pbr.version

from etcd3gw.client import client
from etcd3gw.client import Etcd3Client
from etcd3gw.lease import Lease
from etcd3gw.lock import Lock
from etcd3gw import utils

__version__ = pbr.version.VersionInfo(
    'etcd3gw').version_string()

__all__ = (
    'Etcd3Client',
    'Lease',
    'Lock',
    'client',
    'utils',
)
